@@ -5,6 +5,9 @@ import html
 
 from std.sets import Range, Union
 
+def add(arr, val):
+    return [x + val for x in arr]
+
 class XMLNode:
     is_XMLNodeCaret = False
     is_XMLNodeText = False
@@ -44,6 +47,8 @@ class XMLNode:
     def __str__(self):
         return self.physicalText
 
+    __repr__ = __str__
+
     def append_left_tag(self, tag):
         caret = XMLNodeCaret()
         parent = self.parent
@@ -137,10 +142,6 @@ class XMLNode:
                         style_input[i].add(tag)
         return style_input
     
-    def getPhysicalIndices(self, start, stop):
-        start = self.logical2physicalTagStart(start)
-        stop = self.logical2physicalTagEnd(stop - 1) + 1
-        return start, stop
 
 class XMLNodeCaret(XMLNode):
     is_XMLNodeCaret = True
@@ -236,11 +237,15 @@ class XMLNodeText(XMLNode):
     def physical2logical(self, pos):
         return pos
     
-    def logical2physicalTagStart(self, pos):
-        return pos
+    def is_physically_valid(self, pos):
+        return 0 <= pos < self.length
 
-    logical2physicalTagEnd = logical2physicalTagStart
-    
+    def is_sibling(self, start, end):
+        return 0 <= start < end <= self.length 
+        
+    def getPhysicalIndices(self, start, stop):
+        return start, stop
+        
     def enumerate(self):
         yield (self.lexeme, False)
 
@@ -344,22 +349,38 @@ class XMLNodeBinaryTag(XMLNode):
         return self.arg.logical2physical(pos) + self.tagBegin.length
 
     def physical2logical(self, pos):
-        return clip(self.arg.physical2logical(pos) - self.tagBegin.length, 0, len(self.text) - 1)
+        return clip(self.arg.physical2logical(pos - self.tagBegin.length), 0, len(self.text) - 1)
     
-    def logical2physicalTagStart(self, pos):
-        pos = self.arg.logical2physicalTagStart(pos)
-        if pos:
-            pos += self.tagBegin.length
-        return pos
-    
-    def logical2physicalTagEnd(self, pos):
-        pos = self.arg.logical2physicalTagEnd(pos)
-        if pos:
-            if pos == len(str(self.arg)) - 1:
-                pos += self.tagEnd.length
-            pos += self.tagBegin.length
-        return pos
+    def is_physically_valid(self, pos):
+        if self.tagBegin.length <= pos < self.length - self.tagEnd.length:
+            return self.arg.is_physically_valid(pos - self.tagBegin.length)
     
+    def is_sibling(self, start, end):
+        if self.tagBegin.length <= start < end <= self.length - self.tagEnd.length:
+            start -= self.tagBegin.length
+            end -= self.tagBegin.length
+            return self.arg.is_sibling(start, end)
+
+    def getPhysicalIndices(self, start, stop):
+        start, stop = self.arg.getPhysicalIndices(start, stop)
+        
+        physicalText = str(self.arg)
+        _stop = stop
+        #ignore white spaces to the end;
+        while _stop < len(physicalText) and physicalText[_stop].isspace():
+            _stop += 1
+        
+        if _stop == len(physicalText):
+            if not start:
+                stop = _stop + self.tagEnd.length
+
+        stop += self.tagBegin.length
+        
+        if start:
+            start += self.tagBegin.length
+            
+        return start, stop
+
     def replace(self, old, node):
         assert self.arg is old, self
         self.arg = node
@@ -512,12 +533,15 @@ class XMLNodeSingleTag(XMLNode):
     def physical2logical(self, pos):
         return 0
     
-    def logical2physicalTagStart(self, pos):
-        return 0
-    
-    def logical2physicalTagEnd(self, pos):
-        return self.arg.length - 1
-    
+    def is_physically_valid(self, pos):
+        ...
+            
+    def is_sibling(self, start, end):
+        ...
+            
+    def getPhysicalIndices(self, start, stop):
+        return 0, self.arg.length
+
     def enumerate(self):        
         yield from self.ptr.enumerate()
 
@@ -530,7 +554,7 @@ class XMLNodeSingleTag(XMLNode):
 class XMLNodeArray(XMLNode):
     is_XMLNodeArray = True
     
-    def __init__(self, args, parent):
+    def __init__(self, args, parent=None):
         super().__init__(parent)
         self.args = args
         for ptr in args:
@@ -707,18 +731,58 @@ class XMLNodeArray(XMLNode):
         prev_start = physicalOffset[index][0]
         return self.args[index].physical2logical(pos - prev_start) + prev_start - offsets[index]
     
-    def logical2physicalTagStart(self, pos):
+    def is_physically_valid(self, pos):
+        physicalOffset, offsets = self.physicalOffset, self.offsets
+        assert physicalOffset, physicalOffset
+        index = binary_search(physicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
+        if index < len(physicalOffset):
+            prev_start = physicalOffset[index][0]
+            pos -= prev_start
+            if pos >= 0:
+                return self.args[index].is_physically_valid(pos)
+        
+    def is_sibling(self, start, end):
+        if start < end:
+            physicalOffset, offsets = self.physicalOffset, self.offsets
+            assert physicalOffset, physicalOffset
+            index = binary_search(physicalOffset, start, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
+            _index = binary_search(physicalOffset, end - 1, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
+            if _index == index < len(physicalOffset):
+                prev_start = physicalOffset[index][0]
+                start -= prev_start
+                end -= prev_start
+                if start >= 0:
+                    return self.args[index].is_sibling(start, end)
+                
+            if index < _index < len(physicalOffset):
+                if self.args[index].is_XMLNodeText:
+                    if self.args[_index].is_XMLNodeText:
+                        return True
+                    
+                    if self.args[_index].is_XMLNodeBinaryTag or self.args[_index].is_XMLNodeSingleTag:
+                        prev_start = physicalOffset[_index][0]
+                        end -= prev_start
+                        if end == self.args[_index].length:
+                            return True
+
+    def cmp(self, args, hit):
+        return -1 if hit >= args[1] else 1 if hit < args[0] else 0
+        
+    def getPhysicalIndices(self, start, stop):
         logicalOffset, offsets = self.logicalOffset, self.offsets
-        index = binary_search(logicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
-        prev_start = logicalOffset[index][0]
-        return self.args[index].logical2physicalTagStart(pos - prev_start) + prev_start + offsets[index]
-    
-    def logical2physicalTagEnd(self, pos):
-        logicalOffset, offsets = self.logicalOffset, self.offsets
-        index = binary_search(logicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
-        prev_start = logicalOffset[index][0]
-        return self.args[index].logical2physicalTagEnd(pos - prev_start) + prev_start + offsets[index]
-    
+        index = binary_search(logicalOffset, start, self.cmp)
+        _index = binary_search(logicalOffset, stop - 1, self.cmp)
+        
+        if index == _index:
+            prev_start = logicalOffset[index][0]    
+            return add(self.args[index].getPhysicalIndices(start - prev_start, stop - prev_start), prev_start + offsets[index])
+        else:
+            [prev_start, prev_stop] = logicalOffset[index]
+            _prev_start = logicalOffset[_index][0]    
+            start = add(self.args[index].getPhysicalIndices(start - prev_start, prev_stop - prev_start), prev_start + offsets[index])[0]
+            stop = add(self.args[_index].getPhysicalIndices(0, stop - _prev_start), _prev_start + offsets[_index])[1]
+            return [start, stop]
+
     def sanctity_check(self):
         src = str(self)
         for args in self.interval(''):
@@ -907,13 +971,20 @@ def detect_style(interval, style):
         i += 1
             
     return interval
-    
+
+tagName = '([a-z][-:_a-z]*\d*)'
+attrList = '''(?:\s+:?[a-z][-:_a-z]*=(?:"[^"]*"|'[^']*'))*\s*'''    
+tagBegin = f'<{tagName}{attrList}>'
+tagRegex = fr'''{tagBegin}(?=[\s\S]*?<\/\1>)|</{tagName}>|<(img|mspace|br){attrList}/>|&(#[0-9]+|#x[0-9a-f]+|[^\t\n\f <&#;]{{1,32}});'''
+
+# print("tagRegex =", tagRegex)
+
 def construct_rich_text(text):
     start = 0
     
     richTexts = []
     leftTagCount = {} 
-    for m in re.finditer(r'''<([a-z][-:_a-z]*\d*)(?:\s+:?[a-z][-:_a-z]*=(?:"[^"]*"|'[^']*'))*\s*>(?=[\s\S]*?<\/\1>)|</([a-z][-:_a-z]*\d*)>|<(img|mspace|br)(?:\s+:?[a-z][-:_a-z]*=(?:"[^"]*"|'[^']*'))*\s*/>|&(#[0-9]+|#x[0-9a-f]+|[^\t\n\f <&#;]{1,32});''', text, re.IGNORECASE):
+    for m in re.finditer(tagRegex, text, re.IGNORECASE):
 #                             1-----------------)                                                                     2-----------------)   3-------------)                                                   4----------------------------------------)
         prevText = text[start:m.start()]
         if prevText:
@@ -960,8 +1031,7 @@ class XMLText:
     def __str__(self):
         return self.src[self.start:self.stop]
     
-    def __repr__(self):
-        return str(self)
+    __repr__ = __str__
     
     is_PlainText = False
     is_TagEnd = False
